
prompt
prompt
prompt creation package emp_ext_tab_dml

prompt
prompt
prompt creation package header emp_ext_tab_dml

CREATE OR REPLACE PACKAGE emp_ext_tab_dml
AS

/*********************************************************************
 Public procedure declarations.
*********************************************************************/

  procedure delete_record
  ( p_old_rownumber  IN  number
  );

--

  procedure insert_record
  ( p_new_empno     IN  number
  , p_new_ename     IN  varchar2
  , p_new_job       IN  varchar2
  , p_new_mgr       IN  number
  , p_new_hiredate  IN  date
  , p_new_sal       IN  number
  , p_new_comm      IN  number
  , p_new_deptno    IN  number
  );

--

  procedure update_record
  ( p_old_rownumber  IN  number
  , p_new_empno      IN  number
  , p_new_ename      IN  varchar2
  , p_new_job        IN  varchar2
  , p_new_mgr        IN  number
  , p_new_hiredate   IN  date
  , p_new_sal        IN  number
  , p_new_comm       IN  number
  , p_new_deptno     IN  number
  );

--

END emp_ext_tab_dml;
/

show errors


prompt
prompt
prompt creation package body emp_ext_tab_dml

CREATE OR REPLACE PACKAGE BODY emp_ext_tab_dml
AS

/*********************************************************************
 Private package constants.
*********************************************************************/

  --
  -- The INSTEAD OF triggers on the external table view require
  -- information about which server directory and filenames to use.
  -- In this example code, these are hardcoded. However, they could
  -- be stored in a separate metadata table.
  --
  pc_directory        constant varchar2(256):= 'EXT_TABLES'          ;
  pc_filename_orig    constant varchar2(256):= 'emp_ext_tab.csv'     ;
  pc_filename_temp    constant varchar2(256):= 'emp_ext_tab_temp.csv';

  pc_appendmode       constant varchar2(1)  := 'a';
  pc_readmode         constant varchar2(1)  := 'r';
  pc_writemode        constant varchar2(1)  := 'w';


/*********************************************************************
 Private package cursor.
*********************************************************************/

  --
  -- Note:
  -- Ordering the cursor selection on rownumber asc is essential !
  -- It ensures - in the case of updates - that the pre-statement
  -- order of lines in the file is maintained. This is important
  -- because a multiple update actually gets processed as a series
  -- of single-record updates, *each* of which causes a rewrite of
  -- the external table file. (This also explains why a multiple
  -- delete does not work.)
  --
  cursor c_data
  is
    select   rownumber
           , empno
           , ename
           , job
           , mgr
           , hiredate
           , sal
           , comm
           , deptno
    from     emp_ext_tab
    order by rownumber asc
    ;


/*********************************************************************
 Private procedure definitions.
*********************************************************************/

function get_lock
return integer
is
/*********************************************************************
 This function establishes a user-defined lock. If the lock cannot be
 obtained, an exception is raised to warn the user he should not
 attempt DML at this moment.
 Outputparameter is the lock_id, which has been defined as a constant.
 It has an arbitrary value because we need to lock the table-file, not
 just one record.
 Note that this setup only works if all sessions make use of this
 package. Basically, we only have a user-defined, logical lock.
 Nothing prevents another session from bypassing our locking mechanism
 by using UTL_FILE explicitly to open and write to the external table
 file.
*********************************************************************/
  --
  c_expiration_secs   constant integer := 86400;  -- time-out after 24 hours
  c_lock_id           constant integer := 12345;  -- an arbitrary value !
  c_lock_success      constant integer := 0    ;  --    returncodes:
                                                  --    0 - success
                                                  --    1 - timeout
                                                  --    2 - deadlock
                                                  --    3 - parameter error
                                                  --    4 - already own lock specified by 'id' or 'lockhandle'
                                                  --    5 - illegal lockhandle
  --
  l_request_result             integer;
  --
begin
  --
  l_request_result := dbms_lock.request
                      ( c_lock_id          -- id                in integer
                      , dbms_lock.x_mode   -- lockmode          in integer default x_mode
                      , 0                  -- timeout           in integer default maxwait
                      , FALSE              -- release_on_commit in boolean default FALSE
                      );
  if ( l_request_result <> c_lock_success )
  then
    raise_application_error
    ( -20004
    ,    'EMP_EXT_ATB_DML.get_lock: could not establish lock on external table '
      || '(dbms_lock returncode = '
      || to_char ( l_request_result )
      || ')'
    );
  else
    return c_lock_id;
  end if;
  --
end get_lock;

--

procedure release_lock
( p_lock_id  IN  integer
)
is
/*********************************************************************
 This function releases a user-defined lock.
 It takes the previously established lock_id as an inputparameter.
*********************************************************************/
  --
  l_release_result             integer     ;
  c_release_success   constant integer := 0;  --    returncodes:
                                              --    0 - success
                                              --    3 - parameter error
                                              --    4 - don't own lock specified by 'id' or 'lockhandle'
                                              --    5 - illegal lockhandle
  --
begin
  --
  l_release_result := dbms_lock.release ( p_lock_id );
  --
  if ( l_release_result <> c_release_success )
  then
    raise_application_error
    ( -20005
    ,    'EMP_EXT_ATB_DML.release_lock: could not release lock on external table '
      || '(dbms_lock returncode = '
      || to_char ( l_release_result )
      || ')'
    );
  end if;
  --
end release_lock;

--

function concatenate_data_into_textline
( p_empno     IN  number
, p_ename     IN  varchar2
, p_job       IN  varchar2
, p_mgr       IN  number
, p_hiredate  IN  date
, p_sal       IN  number
, p_comm      IN  number
, p_deptno    IN  number
)
return varchar2
is
/*********************************************************************
 This function returns a string, formatted as a line from the external
 table file. Inputparameters are the columnvalues.
 Within this package, all knowledge of the record layout at the
 filelevel is concentrated within this function. This - obviously -
 includes at least the fieldsequence and basic formatting. However, it
 could also be used for implementing all kinds of constraints
 (null/not null, allowable values, business rules etc.) as well as
 referential integrity.
*********************************************************************/
  --
  c_date_format      constant varchar2(9)  := 'dd-MON-yy';
  c_fieldseparator   constant varchar2(1)  := ','        ;
  --
begin
  --
  return to_char ( p_empno                   ) || c_fieldseparator
      || upper   ( p_ename                   ) || c_fieldseparator
      || upper   ( p_job                     ) || c_fieldseparator
      || to_char ( p_mgr                     ) || c_fieldseparator
      || to_char ( p_hiredate, c_date_format ) || c_fieldseparator
      || to_char ( p_sal                     ) || c_fieldseparator
      || to_char ( p_comm                    ) || c_fieldseparator
      || to_char ( p_deptno                  )
      ;
  --
end concatenate_data_into_textline;

--

procedure close_file_when_open
( p_filehandle  IN OUT  utl_file.file_type
)
is
/*********************************************************************
 This procedure closes a file on the operating system, but only in
 case the file is still open. Inputparameter is the filehandle.
*********************************************************************/
  --
  l_file_is_open   boolean;
  --
begin
  --
  l_file_is_open := utl_file.is_open ( p_filehandle );
  --
  if ( l_file_is_open )
  then
    utl_file.fclose ( p_filehandle );
  end if;
  --
end close_file_when_open;

--

function open_file
( p_filename  IN  varchar2
, p_filemode  IN  varchar2
)
return utl_file.file_type
is
/*********************************************************************
 This function returns a filehandle to a file which is located in the
 server directory. Inputparameters are the filename and the open mode.
*********************************************************************/
begin
  --
  return utl_file.fopen ( pc_directory
                        , p_filename
                        , p_filemode
                        );
  --
end open_file;

--

function open_ext_tab_file
( p_filemode  IN  varchar2
)
return utl_file.file_type
is
/*********************************************************************
 This function returns a filehandle to the external table file.
 Inputparameter is the open mode for the file.
*********************************************************************/
begin
  --
  return open_file ( pc_filename_orig
                   , p_filemode
                   );
  --
end open_ext_tab_file;

--

function open_tmp_tab_file
( p_filemode  IN  varchar2
)
return utl_file.file_type
is
/*********************************************************************
 This function returns a filehandle to the temporary file.
 Inputparameter is the open mode for the file.
*********************************************************************/
begin
  --
  return open_file ( pc_filename_temp
                   , p_filemode
                   );
  --
end open_tmp_tab_file;

--

procedure copy_temp_2_orig_then_del_temp
is
/*********************************************************************
 This procedure first copies the temporary file to the external
 table file, then deletes the temporary file.
*********************************************************************/
begin
  --
  utl_file.fcopy ( pc_directory      -- src_location  IN VARCHAR2
                 , pc_filename_temp  -- src_filename  IN VARCHAR2
                 , pc_directory      -- dest_location IN VARCHAR2
                 , pc_filename_orig  -- dest_filename IN VARCHAR2
                 );
  utl_file.fremove ( pc_directory
                   , pc_filename_temp
                   );
  --
end copy_temp_2_orig_then_del_temp;


/*********************************************************************
 Public procedure definitions.
*********************************************************************/

procedure delete_record
( p_old_rownumber  IN  number
)
is
/*********************************************************************
 This procedure performs the deletion of a record from an external
 table file.
*********************************************************************/
  --
  l_filehandle_orig   utl_file.file_type;
  l_filehandle_temp   utl_file.file_type;
  l_textline          varchar2(1000)    ;
  l_lock_id           integer           ;
  --
begin
  --
  -- Place a lock to prevent concurrent DML actions.
  --
  l_lock_id := get_lock;
  --
  -- Open the external table file for read, and a temporary file for write.
  --
  l_filehandle_orig := open_ext_tab_file ( pc_readmode  );
  l_filehandle_temp := open_tmp_tab_file ( pc_writemode );
  --
  -- Use a cursor-for-loop to check each line in the original file.
  -- The lines that satisfy the condition in the where-clause
  -- of the delete-statement are identified by their rownumber.
  -- Note: INSTEAD OF triggers are implicitly FOR EACH ROW.
  --       In case of a delete, this means only the first
  --       record is processed correctly !
  --
  for r_data in c_data
  loop
    --
    if ( p_old_rownumber <> r_data.rownumber )
    then
      --
      l_textline := concatenate_data_into_textline
                    ( r_data.empno
                    , r_data.ename
                    , r_data.job
                    , r_data.mgr
                    , r_data.hiredate
                    , r_data.sal
                    , r_data.comm
                    , r_data.deptno
                    );
      utl_file.put_line ( l_filehandle_temp
                        , l_textline
                        );
      --
    end if;
    --
  end loop;
  --
  -- Close both files.
  --
  close_file_when_open ( l_filehandle_orig );
  close_file_when_open ( l_filehandle_temp );
  --
  -- Copy the temporary file to the current external table file,
  -- then delete the temporary file.
  --
  copy_temp_2_orig_then_del_temp;
  --
  -- Release the lock.
  --
  release_lock ( l_lock_id );
  --
exception
  --
  when others then
    --
    close_file_when_open ( l_filehandle_orig );
    close_file_when_open ( l_filehandle_temp );
    --
    l_textline := 'Exception in procedure emp_ext_tab_dml.delete_record: '
               || substr(SQLERRM,1,900);
    raise_application_error (-20001, l_textline);
    --
end delete_record;

--

procedure insert_record
( p_new_empno     IN  number
, p_new_ename     IN  varchar2
, p_new_job       IN  varchar2
, p_new_mgr       IN  number
, p_new_hiredate  IN  date
, p_new_sal       IN  number
, p_new_comm      IN  number
, p_new_deptno    IN  number
)
is
/*********************************************************************
 This procedure performs the insert of a record into an external
 table file.
*********************************************************************/
  --
  l_filehandle_orig   utl_file.file_type;
  l_textline          varchar2(1000)    ;
  l_lock_id           integer           ;
  --
begin
  --
  -- Place a lock to prevent concurrent DML actions.
  --
  l_lock_id := get_lock;
  --
  -- Open the external table file for append.
  --
  l_filehandle_orig := open_ext_tab_file ( pc_appendmode );
  --
  -- Concatenate the new fieldvalues into one line of text,
  -- and write the line to file.
  --
  l_textline := concatenate_data_into_textline
                ( p_new_empno
                , p_new_ename
                , p_new_job
                , p_new_mgr
                , p_new_hiredate
                , p_new_sal
                , p_new_comm
                , p_new_deptno
                );
  utl_file.put_line ( l_filehandle_orig
                    , l_textline
                    );
  --
  -- Close the external table file.
  --
  close_file_when_open ( l_filehandle_orig );
  --
  -- Release the lock.
  --
  release_lock ( l_lock_id );
  --
exception
  --
  when others then
    --
    close_file_when_open ( l_filehandle_orig );
    --
    l_textline := 'Exception in procedure emp_ext_tab_dml.insert_record: '
               || substr(SQLERRM,1,900);
    raise_application_error (-20002, l_textline);
    --
end insert_record;

--

procedure update_record
( p_old_rownumber   IN  number
, p_new_empno       IN  number
, p_new_ename       IN  varchar2
, p_new_job         IN  varchar2
, p_new_mgr         IN  number
, p_new_hiredate    IN  date
, p_new_sal         IN  number
, p_new_comm        IN  number
, p_new_deptno      IN  number
)
is
/*********************************************************************
 This procedure performs the update of a record in an external
 table file.
*********************************************************************/
  --
  l_filehandle_orig   utl_file.file_type;
  l_filehandle_temp   utl_file.file_type;
  l_textline          varchar2(1000)    ;
  l_lock_id           integer           ;
  --
begin
  --
  -- Place a lock to prevent concurrent DML actions.
  --
  l_lock_id := get_lock;
  --
  -- Open the external table file for read, and a temporary file for write.
  --
  l_filehandle_orig := open_ext_tab_file ( pc_readmode  );
  l_filehandle_temp := open_tmp_tab_file ( pc_writemode );
  --
  -- Use a cursor-for-loop to check each line in the original file.
  -- The lines that satisfy the condition in the where-clause
  -- of the update-statement are identified by their rownumber.
  -- Note: INSTEAD OF triggers are implicitly FOR EACH ROW.
  --
  for r_data in c_data
  loop
    --
    if ( p_old_rownumber <> r_data.rownumber )
    then
      l_textline := concatenate_data_into_textline
                    ( r_data.empno
                    , r_data.ename
                    , r_data.job
                    , r_data.mgr
                    , r_data.hiredate
                    , r_data.sal
                    , r_data.comm
                    , r_data.deptno
                    );
    else
      l_textline := concatenate_data_into_textline
                    ( p_new_empno
                    , p_new_ename
                    , p_new_job
                    , p_new_mgr
                    , p_new_hiredate
                    , p_new_sal
                    , p_new_comm
                    , p_new_deptno
                    );
    end if;
    --
    utl_file.put_line ( l_filehandle_temp
                      , l_textline
                      );
    --
  end loop;
  --
  -- Close both files.
  --
  close_file_when_open ( l_filehandle_orig );
  close_file_when_open ( l_filehandle_temp );
  --
  -- Copy the temporary file to the current external table file,
  -- then delete the temporary file.
  --
  copy_temp_2_orig_then_del_temp;
  --
  -- Release the lock.
  --
  release_lock ( l_lock_id );
  --
exception
  --
  when others then
    --
    close_file_when_open ( l_filehandle_orig );
    close_file_when_open ( l_filehandle_temp );
    --
    l_textline := 'Exception in procedure emp_ext_tab_dml.update_record: '
               || substr(SQLERRM,1,900);
    raise_application_error (-20003, l_textline);
    --
end update_record;

--

END emp_ext_tab_dml;
/

show errors
